/**********************************************************************************************************************
 * LEDMOTION.cpp
 **********************************************************************************************************************/
/**********************************************************************************************************************
 Copyright (c) 2017 Roel Arits

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 **********************************************************************************************************************/
/*
 y = y0 + (V0 * t) + (0.5 * a * t^2)

 y = distance
 y0 = start distance
 v0 = start velocity
 a = acceleration/deceleration
 t = time

 y  ^                          *
    |                         *
    |                        *
    |                       *
    |                      *
    |                     *
    |                    *
    |                   *
    |                  *
    |                *
    |              *
    |            *
    |          *
    |      *
 y0 * *
    |
    |
    |
 ----+------------------------------------------->
    |                                       time
 */

/***********************************************************************************************************************
 * I N C L U D E S
 **********************************************************************************************************************/
#include "LEDMotion.h"

/***********************************************************************************************************************
 * L O C A L   S Y M B O L   D E F I N I T I O N S (defines, typedefs)
 **********************************************************************************************************************/
#define ANGLE_THRESHOLD                   (0.2F) // angle should be above this threshold to trigger an event

#define BOUNCE_DECEL_FACTOR               (0.7F) // Increase to increase deceleration during bouncing
#define BOUNCE_VELOCITY_DAMPING           (0.7F) // the higher the value, the less damping

#define BRAKING_DECEL_FACTOR              (0.15F)  // Increase to increase deceleration

#define MIN_ACCELERATION                  (2.0F) // Increase to increase the base acceleration. Don't use 0
#define MAX_ACCELERATION                 (100.0F) // Increase to increase max acceleration

#define MIN_PITCH                         (0.0F)  // lowest pitch angle
#define MAX_PITCH                        (45.0F) // highest pitch angle

#define LED_COLOR_BACKGROUND              0x000004
#define LED_COLOR_BALL                    0x7F007F

#define LED_COLOR_BALL_MOVING             LED_COLOR_BALL
#define LED_COLOR_BALL_DECELERATING       LED_COLOR_BALL
#define LED_COLOR_BALL_STOPPED            LED_COLOR_BALL

#define DEBUG_COUNT_MAX                   50

/***********************************************************************************************************************
 * D A T A   D E C L A R A T I O N S (exported, local)
 **********************************************************************************************************************/
// We don't want to throw out debug information at high speed in the main loop,
// so we need to slow this down using a counter and a max count = DEBUG_COUNT_MAX.
uint32_t debugCount = 0;

/***********************************************************************************************************************
 * L O C A L   F U N C T I O N S
 **********************************************************************************************************************/

/***********************************************************************************************************************
 * E X P O R T E D   F U N C T I O N S
 **********************************************************************************************************************/
//---------------------------------------------------------------------------------------
// Constructor
//---------------------------------------------------------------------------------------
TLEDMotion::TLEDMotion ()
{
   startTime = 0L;

   startLEDNr = 0;
   LEDNr = 0;
   prevLEDNr = 0;

   dir = DIRECTION_NONE;
   prevDir = dir;
   decelerationDir = dir;

   distance = 0.0F;
   prevDistance = distance;
   storedDistance = distance;

   event = MOTION_EVENT_NONE;
   prevEvent = MOTION_EVENT_NONE;

   state = MOTION_STATE_READY;
   newState = state;

   velocity = 0.0F;
   prevVelocity = velocity;
   storedVelocity = velocity;

   acceleration = 0.0F;
   deceleration = 0.0F;

   LEDColor = LED_COLOR_BALL_MOVING;
}

//---------------------------------------------------------------------------------------
// Destructor
//---------------------------------------------------------------------------------------
TLEDMotion::~TLEDMotion ()
{

}

//---------------------------------------------------------------------------------------
// Convert angle from MPU650 into an acceleration, direction and an event
//---------------------------------------------------------------------------------------
void
TLEDMotion::ConvertAngleToAccelearation (Adafruit_NeoPixel *pStrip, float argAngle)
{
   // Saturate the pitch to the MAX_PITCH, so acceleration calculation is always valid
   if (fabs (argAngle) > MAX_PITCH)
   {
      argAngle = MAX_PITCH;
   }

   float newAcceleration = MIN_ACCELERATION + ((MAX_ACCELERATION - MIN_ACCELERATION) * fabs (argAngle) / (MAX_PITCH - MIN_PITCH));

   // The acceleration should only increase when moving into a certain direction
   // When the acceleration would decrease while moving into a certain direction, this would mean that the velocity becomes negative, so the distance would decrease instead of increase.
// // acceleration is set to 0 when 
   if (newAcceleration >= acceleration)
   {
      acceleration = newAcceleration;
   }
}

//---------------------------------------------------------------------------------------
// Move LED positive
//---------------------------------------------------------------------------------------
void
TLEDMotion::MoveLEDPos (Adafruit_NeoPixel *pStrip)
{
   if ((startLEDNr + distance) > (pStrip->numPixels() - 1))
   {
      LEDNr = pStrip->numPixels() - 1;
      startLEDNr = LEDNr;

      if (state == MOTION_STATE_STOPPED_AT_END)
      {
         // Start new bounce with decreased velocity
         storedVelocity = BOUNCE_VELOCITY_DAMPING * storedVelocity;
      }
      else
      {
         storedVelocity = velocity;
      }

      newState = MOTION_STATE_STOPPED_AT_END;
      startTime = millis ();
   }
   else
   {
      LEDNr = startLEDNr + distance;  
   }
}

//---------------------------------------------------------------------------------------
// Move LED negative
//---------------------------------------------------------------------------------------
void
TLEDMotion::MoveLEDNeg (Adafruit_NeoPixel *pStrip)
{
   if ((startLEDNr - distance) < 0)
   {
      LEDNr = 0;
      startLEDNr = LEDNr;

      if (state == MOTION_STATE_STOPPED_AT_START)
      {
         // Start new bounce with decreased velocity
         storedVelocity = BOUNCE_VELOCITY_DAMPING * storedVelocity;
      }
      else
      {
         storedVelocity = velocity;
      }

      newState = MOTION_STATE_STOPPED_AT_START;
      startTime = millis ();
   }
   else
   {
      LEDNr = startLEDNr - distance;
   }
}

//---------------------------------------------------------------------------------------
// Prepare start
//---------------------------------------------------------------------------------------
void
TLEDMotion::PrepareStart (Adafruit_NeoPixel *pStrip)
{
   startLEDNr = (pStrip->numPixels () / 2);
   LEDNr = startLEDNr;
   prevLEDNr = startLEDNr;

   for (uint8_t i = 0; i < pStrip->numPixels (); i++)
   {
      pStrip->setPixelColor (i, LED_COLOR_BACKGROUND);
   }

   pStrip->setPixelColor (startLEDNr, LED_COLOR_BALL_MOVING);
   pStrip->show ();
}

//---------------------------------------------------------------------------------------
// Prepare deceleration
//---------------------------------------------------------------------------------------
void
TLEDMotion::PrepareDeceleration (void)
{
   storedDistance = prevDistance;
   storedVelocity = prevVelocity;
   deceleration = BRAKING_DECEL_FACTOR * acceleration;
   startTime = millis ();
}

//---------------------------------------------------------------------------------------
// Handle and clear event
//---------------------------------------------------------------------------------------
void
TLEDMotion::HandleEvent (void)
{
   if (event != MOTION_EVENT_NONE)
   {
#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION)  || (MOTION_DEBUG == MOTION_DEBUG_STATES))
      if (event != prevEvent)
      {
         switch(event)
         {
            case MOTION_EVENT_MOVE:             Serial.println ("eM"); break;
            case MOTION_EVENT_CHANGE_DIR:       Serial.println ("eD"); break;
            case MOTION_EVENT_STOP:             Serial.println ("eS"); break;
         }
         prevEvent = event;
      }
#endif

      switch (event)
      {
         case MOTION_EVENT_MOVE:
         {
            switch (state)
            {
               case MOTION_STATE_READY:
               {
                  // Start moving
                  velocity = 0.0F;
                  prevVelocity = 0.0F;
                  acceleration = 0.0F;

                  storedVelocity = 0.0F;
                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;

               case MOTION_STATE_DECEL_STOPPED:
               {
                  // Start moving from stop or from decelerating
                  velocity = 0.0F;
                  prevVelocity = 0.0F;
                  
                  startLEDNr = prevLEDNr;
                  
                  acceleration = 0.0F;
                  
                  storedVelocity = 0.0F;

                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;

               case MOTION_STATE_STOPPED_AT_START:
               case MOTION_STATE_STOPPED_AT_END:
               {
                  // Start moving from stop or from decelerating
                  velocity = 0.0F;
                  prevVelocity = 0.0F;

                  startLEDNr = LEDNr;

                  acceleration = 0.0F;
                  storedVelocity = 0.0F;
                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;

               case MOTION_STATE_DECELERATING:
               {
                  prevDir = dir;

                  startLEDNr = LEDNr;
                  
                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;
            }
         }
         break;

         case MOTION_EVENT_CHANGE_DIR:
         {
            switch (state)
            {
               case MOTION_STATE_STOPPED_AT_START:
               case MOTION_STATE_STOPPED_AT_END:
               case MOTION_STATE_DECEL_STOPPED:
               {
                  // Change direction from stop
                  startLEDNr = LEDNr;
                  
                  //acceleration = 0.0F;
                  //velocity = 0.0F;  //NEEDED ?
                  //storedVelocity = 0.0F; //NEEDED ?
                  
                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;

               case MOTION_STATE_MOVING:
               {
                  // When direction is changed while moving, then decelerate
                  PrepareDeceleration ();
                  newState = MOTION_STATE_DECELERATING;
               }
               break;

               /*case MOTION_STATE_DECELERATING:
               {
                  startLEDNr = LEDNr;
                  newState = MOTION_STATE_MOVING;
                  startTime = millis ();
               }
               break;*/
            }
         }
         break;

         case MOTION_EVENT_STOP:
         {
            switch (state)
            {
               case MOTION_STATE_MOVING:
               {
                  // Stop during moving, so decelerate
                  PrepareDeceleration ();
                  newState = MOTION_STATE_DECELERATING;
               }
               break;
            }
         }
         break;

      }
      event = MOTION_EVENT_NONE;
   }
}

//---------------------------------------------------------------------------------------
// Handle a state change
//---------------------------------------------------------------------------------------
void
TLEDMotion::HandleStateChange (void)
{
   if (newState != state)
   {
      state = newState;
#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION)  || (MOTION_DEBUG == MOTION_DEBUG_STATES))
      switch(state)
      {
         case MOTION_STATE_READY:            Serial.println ("R"); break;
         case MOTION_STATE_MOVING:           Serial.println ("M"); break;
         case MOTION_STATE_CHANGING_DIR:     Serial.println ("D"); break;
         case MOTION_STATE_DECELERATING:     Serial.println ("DEC"); break;
         case MOTION_STATE_DECEL_STOPPED:    Serial.println ("SDEC"); break;
         case MOTION_STATE_STOPPED_AT_END:   Serial.println ("Send"); break;
         case MOTION_STATE_STOPPED_AT_START: Serial.println ("Sstr"); break;
      }
#endif
   }
}

//---------------------------------------------------------------------------------------
// Handle start moving from stop
//---------------------------------------------------------------------------------------
void
TLEDMotion::HandleStartMovingState(Adafruit_NeoPixel *pStrip, float elapsedTime)
{
   // Distance is always positive
   // y = y0 + (v0 * t) + (0.5 * a * t^2)
   // When starting from stop, the start distance y0 = 0 and the start velocity V0 = 0
   // So we get y = (0.5 * a * t^2)
   distance = (storedVelocity * elapsedTime) + (0.5F * acceleration * pow (elapsedTime, 2));

   // Calculate velocity using distance and time
   if (fabs(elapsedTime) > 0.001F)
   {
      velocity = (0.5F * distance) / elapsedTime;
   }

#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION))
   if (debugCount > DEBUG_COUNT_MAX)
   {
      Serial.println ((String) "");
      //Serial.println ((String) "time= " + elapsedTime);
      //Serial.println ((String) "acc=  " + acceleration);
      //Serial.println ((String) "vel=  " + velocity);
      Serial.println ((String) "dist= " + distance);
   }
#endif
   switch (dir)
   {
      case DIRECTION_POSITIVE:
      {
         MoveLEDPos (pStrip);
      }
      break;

      case DIRECTION_NEGATIVE:
      {
         MoveLEDNeg (pStrip);
      }
      break;

      // No direction while moving, so decelerate to a stop
      // Check the stored direction at the moment deceleration was started, to decide
      // what direction we are going.
      case DIRECTION_NONE:
      {
         switch (decelerationDir)
         {
            case DIRECTION_POSITIVE:
            {
               MoveLEDPos (pStrip);
            }
            break;
            
            case DIRECTION_NEGATIVE:
            {
               MoveLEDNeg (pStrip);
            }
            break;
         }
      }
      break;
   }
   LEDColor = LED_COLOR_BALL_MOVING;
}

//---------------------------------------------------------------------------------------
// Handle Deceleration (braking) state
//---------------------------------------------------------------------------------------
void
TLEDMotion::HandleDeceleratingState(Adafruit_NeoPixel *pStrip, float elapsedTime)
{
   deceleration = BRAKING_DECEL_FACTOR * acceleration;

   // Distance is always positive
   // Deceleration formula
   // y = y0 + (v0 * t) - (0.5 * a * t^2)
   // y0 = distance at the moment that deceleration is started
   // v0 = velocity at the moment that deceleration is started
   // At a certain moment the distance delta will become negative.
   // At that moment the motion is stopped.
   distance = storedDistance + ((storedVelocity * elapsedTime) - (0.5F * deceleration * pow (elapsedTime, 2)));

   if (fabs(elapsedTime) > 0.001F)
   {
      velocity = (0.5F * distance) / elapsedTime;
   }

#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION))
   if (debugCount > DEBUG_COUNT_MAX)
   {
      Serial.println ((String) "");
      //Serial.println ((String) "time=  " + elapsedTime);
      //Serial.println ((String) "stDst= " + storedDistance);
      //Serial.println ((String) "stVel= " + storedVelocity);
      //Serial.println ((String) "decel= " + deceleration);
      //Serial.println ((String) "prDst= " + prevDistance);
      Serial.println ((String) "dist=  " + distance);
   }
#endif

   // Check the direction that was stored at the moment that deceleration started
   // Keep moving into this stored direction
   switch (decelerationDir)
   {
      case DIRECTION_POSITIVE:
      {
         MoveLEDPos (pStrip);
      }
      break;
      
      case DIRECTION_NEGATIVE:
      {
         MoveLEDNeg (pStrip);
      }
      break;
   }
   LEDColor = LED_COLOR_BALL_DECELERATING;

   // When the distance delta becomes negative, the deceleration is completed
   if ((distance - prevDistance) < 0.0F)
   {
      newState = MOTION_STATE_DECEL_STOPPED;
   }
}

//---------------------------------------------------------------------------------------
// Handle Bouncing state
//---------------------------------------------------------------------------------------
void
TLEDMotion::HandleBouncingState(Adafruit_NeoPixel *pStrip, float elapsedTime)
{
   // distance can be negative
   distance = (storedVelocity * elapsedTime) - (0.5F * deceleration * pow (elapsedTime, 2.0F));

#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION))
   if (debugCount > DEBUG_COUNT_MAX)
   {
      Serial.println ((String) "");
      //Serial.println ((String) "time=  " + elapsedTime);
      //Serial.println ((String) "stVel= " + storedVelocity);
      //Serial.println ((String) "decel= " + deceleration);
      Serial.println ((String) "dist=  " + distance);
   }
#endif

   // When distance becomes negative, the current bounce is completed and we start the next bounce
   if (distance < 0.0F)
   {
      distance = 0.0F;

      // Start new bounce with decreased velocity
      storedVelocity = BOUNCE_VELOCITY_DAMPING * storedVelocity;

      startTime = millis ();
   }

   deceleration = BOUNCE_DECEL_FACTOR * acceleration;

   // Check against which end we are bouncing
   switch (state)
   {
      case MOTION_STATE_STOPPED_AT_START:
      {
         MoveLEDPos(pStrip);
      }
      break;

      case MOTION_STATE_STOPPED_AT_END:
      {
         MoveLEDNeg(pStrip);
      }
      break;
   }
   LEDColor = LED_COLOR_BALL_STOPPED;
}

//---------------------------------------------------------------------------------------
// Update direction
//---------------------------------------------------------------------------------------
void
TLEDMotion::UpdateDirection (float pitch)
{
   // Angle bigger than threshold, then move to end of strip
   if (pitch > ANGLE_THRESHOLD)
   {
      dir = DIRECTION_POSITIVE;
   }
   else if (pitch < -ANGLE_THRESHOLD)
   {
      dir = DIRECTION_NEGATIVE;
   }
   // Below threshold
   else
   {
      dir = DIRECTION_NONE;
   }
}

//---------------------------------------------------------------------------------------
// Change in direction detected, so check if we need to decelerate
//---------------------------------------------------------------------------------------
void
TLEDMotion::CheckForDirectionChange (float pitch)
{
   if (state == MOTION_STATE_DECELERATING)
   {
      if (dir == decelerationDir)
      {
         event = MOTION_EVENT_MOVE;
      }
	   // Don't check and update direction when decelerating
	   return;
   }
   else if (state == MOTION_STATE_DECEL_STOPPED)
   {
       //Check if the direction has changed during deceleration
       // In that case, when stopped, issue a move event;
       if (decelerationDir != dir)
       {
           event = MOTION_EVENT_MOVE;
           return;
       }
   }
   else if (state == MOTION_STATE_STOPPED_AT_END)
   {
      if (dir == DIRECTION_NEGATIVE)
      {
         event = MOTION_EVENT_MOVE;
         return;
      }
   }
   else if (state == MOTION_STATE_STOPPED_AT_START)
   {
      if (dir == DIRECTION_POSITIVE)
      {
         event = MOTION_EVENT_MOVE;
         return;
      }
   }

   // Change in direction detected
   if (dir != prevDir)
   {
      // start moving from stop or going through the stop state while changing direction
      if (prevDir == DIRECTION_NONE)
      {
         event = MOTION_EVENT_MOVE;
      }
      // Change while moving
      else
      {
         // Stop while moving
         if (dir == DIRECTION_NONE)
         {
            event = MOTION_EVENT_STOP;
         }
         // Change dir while moving
         else
         {
            event = MOTION_EVENT_CHANGE_DIR;
         }
      }
      // Store the previous direction
      // This stored direction is used during decelerating
      decelerationDir = prevDir;
      prevDir = dir;
   }
}

//---------------------------------------------------------------------------------------
// Convert a byte to a color
//---------------------------------------------------------------------------------------
uint32_t
TLEDMotion::ConvertByteToColor (byte data, byte max_data)
{
   uint32_t retValue = 0;

   if (data > max_data)
   {
      data = max_data;
   }

   if (data > (5 * max_data) / 7)
   {
      retValue = 0x00FFFF00 + ((((data - (5 * max_data) / 7)) * 3L));
   }
   else if (data > (4 * max_data) / 7)
   {
      retValue = 0x00FF0000 + ((((data - (4 * max_data) / 7)) * 3L) << 8);
   }
   else if (data > (3 * max_data) / 7)
   {
      retValue = 0x00FF0000 + ((((data - (3 * max_data) / 7)) * 3L));
   }
   else if (data > (2 * max_data) / 7)
   {
      retValue = 0x00000000 + ((((data - (2 * max_data) / 7)) * 3L) << 16);
   }
   else if (data > (1 * max_data) / 7)
   {
      retValue = 0x0000FF00 + ((((data - (1 * max_data) / 7)) * 3L));
   }
   else if (data > (0 * max_data) / 7)
   {
      retValue = 0x000000FF + ((((data - (0 * max_data) / 7)) * 3L) << 8);
   }
   return retValue;
}

//---------------------------------------------------------------------------------------
// Main state machine
//---------------------------------------------------------------------------------------
void
TLEDMotion::Main (Adafruit_NeoPixel *pStrip, float yawn, float pitch, float roll)
{
   // Only use the pitch value to measure the angle and convert is to a new acceleration
   ConvertAngleToAccelearation(pStrip, pitch);

   UpdateDirection(pitch);
   
   CheckForDirectionChange(pitch);

   HandleEvent();

   // Calculate elapsed time after handling an event, because in the event handling the start time can be changed
   float elapsedTime = (millis () - startTime) / 1000.0F;

   HandleStateChange();

   //========================= HANDLE STATES =======================================
   switch (state)
   {
      // Start moving from stop
      case MOTION_STATE_MOVING:
      {
         HandleStartMovingState(pStrip, elapsedTime);
      }
      break;

      // Decelerate when direction changes
      case MOTION_STATE_DECELERATING:
      {
         HandleDeceleratingState(pStrip, elapsedTime);
      }
      break;

      // Bounce
      case MOTION_STATE_STOPPED_AT_START:
      case MOTION_STATE_STOPPED_AT_END:
      {
         HandleBouncingState(pStrip, elapsedTime);
      }
      break;

      default:
      {

      }
      break;
   }

   prevDistance = distance;
   prevVelocity = velocity;

   //Only update LED position when there is a change in position
   if (LEDNr != prevLEDNr)
   {
      pStrip->setPixelColor (prevLEDNr, LED_COLOR_BACKGROUND);
      pStrip->setPixelColor (LEDNr, LEDColor);
      //pStrip->setPixelColor (LEDNr, ConvertByteToColor(acceleration, MAX_ACCELERATION));
      pStrip->show ();

      prevLEDNr = LEDNr;
   }
  
#if ((MOTION_DEBUG == MOTION_DEBUG_ALL) || (MOTION_DEBUG == MOTION_DEBUG_MOTION))
   if (debugCount > DEBUG_COUNT_MAX)
   {
      //Serial.println ((String) "acc=   " + acceleration);
      //Serial.println ((String) "pitch= " + pitch);
      //Serial.println ((String) "yawn=  " + yawn);
      //Serial.println ((String) "roll=  " + roll);
      Serial.println ((String) "dir=   " + dir);
      Serial.println ((String) "StrtLED = " + startLEDNr);
      Serial.println ((String) "LEDNr= " + LEDNr);
   }
#elif (MOTION_DEBUG == MOTION_DEBUG_STATES)
   if (debugCount > DEBUG_COUNT_MAX)
   {
      Serial.println ((String) "LEDNr= " + LEDNr + (String) " - StrtLED = " + startLEDNr  + (String) " - Dist = " + distance + (String) " - acc = " + acceleration + (String) " - vel = " + velocity + (String) " - time = " + elapsedTime);
   }
#endif

   HandleStateChange();

   //================================= END =========================================
   if (debugCount > DEBUG_COUNT_MAX)
   {
      debugCount = 0;
   }
   else
   {
      debugCount++;
   }
}
